/*
 * STREAM - Functions to support zero-copy input streams based on fastrings
 *
 * There are two functions to pull from the stream: strm_getline, which
 * returns a p, l string pointing into the stream's ring buffer containing one 
 * newline-separated line, and strm_ensure, which tries to ensure at least
 * the given number of bytes are available, and returns that number, or less,
 * in case of EOF. Note that we only support blocking file descriptors; for
 * non-blocking cases, the decision when to attempt to pull should have little
 * to do with userspace buffering.
 *
 * When data is needed, the stream tries to pull as much data from its file 
 * descriptor as the there is room in its ring buffer. 
 *
 * The given size of the buffer is always rounded up to the nearest page 
 * boundary (4k or 8k for most systems).
 *
 * After you are done with the data given by strm_getline or strm_ensure,
 * you must call strm_advance to advance the read pointer.
 *
 * Author:
 * Emile van Bergen, emile@evbergen.xs4all.nl
 *
 * Permission to redistribute an original or modified version of this program
 * in source, intermediate or object code form is hereby granted exclusively
 * under the terms of the GNU General Public License, version 2. Please see the
 * file COPYING for details, or refer to http://www.gnu.org/copyleft/gpl.html.
 *
 * History:
 * 2004/01/13 - EvB - Created
 */


#include <evblib/fastring/fastring.h>
#include <evblib/str/str.h>

#define strm_advance(strm, amount)	ring_get(&(strm)->r, (amount))
#define strm_attach(strm, newfd)	((strm)->eof = 0, (strm)->fd = (newfd))


typedef struct stream {
    RING_T r;
    int fd, eof; 
    int aux;			/* line counter or other state info */
    size_t bufsize;
} STRM_T;


STRM_T *strm_new(int fd, ssize_t bufsize);
int strm_init(STRM_T *strm, int fd, ssize_t bufsize);
void strm_done(STRM_T *strm);
void strm_del(STRM_T *strm);

STR_T strm_getline(STRM_T *strm);
STR_T strm_ensure(STRM_T *strm, ssize_t len);

