/* UTF-8 and UVL encoding and decoding */

/* UVL, or Unicode Variable Length encoding is the generic name I devised for
 * the encoding scheme used in UTF-8, which can convey any value of 7 to 36 
 * bits in 1-7 octets.
 * 
 * Some background: I distinguish two layers in UTF-8: firstly the variable 
 * length encoding scheme, including the non-overlap requirement (no 
 * overlength encodings allowed), and secondly the assignment of valid unicode 
 * points.
 *
 * The standard joins these two layers and considers everything that doesn't 
 * lead to a defined Unicode code point as ill-formed as anything else, but 
 * for the purpose of this decoder these layers are separated, because I think
 * the first layer (UVL) has wider applicability than just transporting 
 * Unicode.
 *
 * The non-overlap requirement, which makes that each value can be encoded in
 * only one way, is actually much more interesting than the allowed character 
 * range. It gives the encoding its security properties; it is also anticipated
 * that this feature will live through more Unicode revisions than the 1 
 * million character limit. Hence our inclusion of that requirement in layer 1 
 * instead of in layer 2, while Unicode historically seems to have considered 
 * non-overlap less important than valid values: in Unicode revisions before 
 * 4.0, overlap was merely considered problematic, not illegal.
 * 
 * Overlapping values, ie. overlength encodings, are thus flagged as real 
 * encoding errors here, both in the table and in all decoders, resulting in 
 * the full backtracking exercise as specified in Unicode version 4.
 */

#include <evblib/misc/misc.h>	    /* for safe_malloc */

#include <evblib/unicode/utf8.h>    /* for checking our interface prototypes */


/* Table indexed by the first octet of an UVL or UTF-8 sequence, with bits 0-3
 * containing the number of octets to follow and bits 4-7 specifying an error
 * value: bit 7 is set if the octet cannot introduce any valid UVL or UTF-8
 * sequence; all other error bits are undefined in that case. Bit 6 is set if
 * the octet would introduce a value encoded with an overlength sequence. Bit 5
 * is set if the octet introduces an UVL sequence that leads to a value outside 
 * the range of valid unicode code points and therefore does not introduce 
 * a valid UTF-8 sequence. Bit 4 is reserved. */

static const unsigned char uc1[256] = {
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,

    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,

    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,

    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,	    /* 0-7f */

    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,
    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,
    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,
    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,	    

    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,
    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,
    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,
    0xf0, 0xf0, 0xf0, 0xf0,	0xf0, 0xf0, 0xf0, 0xf0,	    /* 80-bf = err */

    0x40, 0x40, 0x01, 0x01,	0x01, 0x01, 0x01, 0x01,	    /* c0,c1 = overl. */
    0x01, 0x01, 0x01, 0x01,	0x01, 0x01, 0x01, 0x01,
    0x01, 0x01, 0x01, 0x01,	0x01, 0x01, 0x01, 0x01,
    0x01, 0x01, 0x01, 0x01,	0x01, 0x01, 0x01, 0x01,	    /* c0-df */

    0x02, 0x02, 0x02, 0x02,	0x02, 0x02, 0x02, 0x02,	
    0x02, 0x02, 0x02, 0x02,	0x02, 0x02, 0x02, 0x02,	    /* e0-ef */

    0x03, 0x03, 0x03, 0x03,	0x03, 0x23, 0x23, 0x23,	    /* f0-f7 */
    0x24, 0x24, 0x24, 0x24,	0x25, 0x25, 0x26, 0xf0,	    /* f8-ff */
};


/*
 * Table indexed by the number of bytes after the first one in a UVL (UTF-8
 * variable length) sequence, containing the bitmask and maximum value for a
 * sequence with that length. It is used both for masking out the introducing
 * one bits and for checking overlap; see utf8dec.inc. Theoretically n=6 allows
 * 36 bits instead of 32, but I don't want to introduce long long dependencies.
 */


static const u_int32_t uchmask[7] = {
    0x7f, 0x7ff, 0xffff, 0x1fffff, 0x3ffffff, 0x7fffffff, 0xffffffff
};


/*************************************************** decoding routines */

/* utf8to16be
 *
 * This function converts an UTF-8 string to an octet string containing 
 * 16-bit big endian values, and returns the number of bytes put in the
 * destination buffer, which may be less than the space provided, since
 * we only output full values. In addition, the number of octets taken
 * from the input buffer is returned in *ret_taken if the pointer is
 * not null.
 *
 * We continue after errors, until all source octets are processed or the
 * destination buffer is full, inserting U+FFFD in the destination stream 
 * for each source octet that contains an error (either layer 1 or 2, as 
 * per the Unicode spec; that is, code point range errors also lead to 
 * backtracking here).
 *
 * Note: this does not convert UTF-8 to UTF-16BE, but UTF-8 to big endian
 * UCS-2, with no surrogate code points allowed. This is incomplete according
 * to the specification, but it's very useful when using X11's built in 16-bit 
 * text service, which cannot deal with more than 65536 characters per face 
 * anyway. 
 */

#define EMIT(ch)    				\
	if (dl-- < 2) break;			\
	*d++ = (unsigned char)(ch >> 8);	\
	*d++ = (unsigned char)ch;		\
	ret++;

#define EMITERR	    EMIT(0xfffd)

#define CHECKRANGE

int utf8to16be(char *s, int sl, char *d, int dl, int *ret_taken)

#include <unicode/utf8dec.inc>

#undef EMIT
#undef EMITERR
#undef CHECKRANGE

/* utf8to32ho
 *
 * convert an utf-8 string into an array of 32-bit unicode code points in host 
 * order. Returns the number of values placed in the destination array; 
 * optionally returns the number of bytes taken from the input buffer in 
 * *ret_taken. */

#define EMIT(ch)    		\
	if (dc-- <= 0) break;	\
	*d++ = ch; 		\
	ret++;

#define EMITERR     		\
	EMIT(0xfffffffd);

#define CHECKRANGE

int utf8to32ho(char *s, int sl, uint32_t *d, int dc, int *ret_taken)

#include <unicode/utf8dec.inc>

#undef EMIT
#undef EMITERR
#undef CHECKRANGE


/* uvlto32ho
 *
 * convert an UVL string into an array of 32-bit values in host order. 
 * Returns the number of values placed in the destination array; optionally 
 * returns the number of bytes taken from the input buffer in *ret_taken. 
 * We stop when the destination buffer is full or when an error occurs. */

#define EMIT(ch)    		\
	if (dc-- <= 0) break;	\
	*d++ = ch; 		\
	ret++;

#define EMITERR     		\
	break;

int uvlto32ho(char *s, int sl, uint32_t *d, int dc, int *ret_taken)

#include <unicode/utf8dec.inc>

#undef EMIT
#undef EMITERR


/* getuvl
 *
 * This routine returns a single value from an UVL string and optionally returns
 * the number of bytes taken from the string in *ret_taken. */

#define EMIT(ch) 	ret = ch; break;
#define EMITERR		break;

int32_t getuvl(char *s, int sl, int *ret_taken)

#include <evblib/unicode/utf8dec.inc>

#undef EMIT
#undef EMITERR


/* utf8len
 *
 * This routine returns the number of unicode characters in a UTF-8 string;
 * we stop at encoding errors. */

#define EMIT(ch)    ret++
#define EMITERR     break
#define CHECKRANGE

int utf8len(char *s, int sl, int *ret_taken)

#include <unicode/utf8dec.inc>

#undef EMIT
#undef EMITERR
#undef CHECKRANGE

/* utf8to16bestr
 *
 * This routine converts UTF-8 to a 16-bit subset of Unicode expressed as big
 * endian values in a standard STR_T structure, where l is the number of octets
 * at p. The memory pointed to by the p member must be freed after use. */

STR_T utf8to16bestr(char *p, int l)
{
    STR_T ret;

    ret.l = utf8len(p, l, 0) << 1;
    if (ret.l) {
	ret.p = (char *)safe_malloc(ret.l);
	utf8to16be(p, l, ret.p, ret.l, 0);
	return ret;
    }
    ret.p = 0;
    return ret;
}


/*************************************************** encoding routines */

/* utf8from32ho
 * 
 * This converts an array of 32-bit values to an UVL string (or UTF-8 string 
 * if the values are within in unicode code point range) in a buffer given by 
 * d, dl; returns the length of the resulting string in octets and optionally 
 * the number of elements left in the source array in *ret_sc_left. */

#define UTEST	(sc > 0)
#define UGET	(*s)
#define UADV	(void)(sc--, *s++)
#define UDONE	(void)(ret_sc_left && (*ret_sc_left = sc))

int utf8from32ho(char *d, int dl, uint32_t *s, int sc, int *ret_sc_left)

#include <unicode/utf8enc.inc>

#undef UTEST
#undef UGET
#undef UADV
#undef UDONE


/* 
 * vim:softtabstop=4:sw=4
 */

