
/* This is the UVL/UTF-8 decoder routine body. You need to prepend a function
 * declaration that contains at least the following arguments:
 *
 *	int utf8towhatever(char *s, int sl, int *ret_taken)
 *
 * and you need to define two macros, EMIT(ch) where ch is an u_int32_t
 * containing the unicode value, and EMITERR. These macros may set or 
 * update ret and use any other argument of your function to perform the 
 * destination side of the conversion; they may also use 'break' to stop 
 * decoding. 
 *
 * If you define CHECKRANGE, sequences leading to undefined unicode values 
 * are treated as encoding errors; this is required for UTF-8, but must be 
 * cleared when generating a layer 1-only decoder (UVL decoder).
 *
 * If the caller passes a non-null pointer, *ret_taken will be set to the
 * number of octets taken from the input buffer (always either nothing or
 * a full UVL sequences)
 *
 * This routine needs the uc1 and uchmask tables from top level (static or 
 * global).
 */

{
    unsigned char *i = (unsigned char *)s;
    unsigned char *e = (unsigned char *)s + sl;
    u_int32_t ch;
    int ret, n, err;

    ret = 0;
    while(i < e) {		/* one iteration per source sequence */

	/* set accumulator ch, get number of octets to follow first one 
 	   and error flag */

	ch = *i++; n = uc1[ch]; 
#ifdef CHECKRANGE
	err = n >> 4;		/* test bits 4-7 */
#else
	err = n >> 6;		/* tests bits 6-7 */
#endif
	n &= 0xf;

	/* if first octet is out of range or no full sequence left in buffer,
	   output an error and go to the next octet if we're not breaking. */

	if (err || i + n > e) { EMITERR; continue; }

	/* get the rest of the sequence; note side effect i += n */

	switch(n) {
	    case 6: err |= (*i & 0xc0) ^ 0x80; ch <<= 6; ch |= *i++ & 0x3f;
	    case 5: err |= (*i & 0xc0) ^ 0x80; ch <<= 6; ch |= *i++ & 0x3f;
	    case 4: err |= (*i & 0xc0) ^ 0x80; ch <<= 6; ch |= *i++ & 0x3f;
	    case 3: err |= (*i & 0xc0) ^ 0x80; ch <<= 6; ch |= *i++ & 0x3f;
	    case 2: err |= (*i & 0xc0) ^ 0x80; ch <<= 6; ch |= *i++ & 0x3f;
	    case 1: err |= (*i & 0xc0) ^ 0x80; ch <<= 6; ch |= *i++ & 0x3f;
	}

	/* handle errors in the sequence by going back to its very first octet,
	   emitting an error value (or breaking), and then skipping that first
	   octet before retrying. */

	if (err) { 
		i -= n + 1; 
		EMITERR; 
		i++; 
		continue; 
	}

	/* the number of bits we want in ch is n * 5 + 7, so the number
	 * of places we need to shift 32 ones to the right is 32 - wantedbits,
	 * or 25 - n * 5, or 25 - n * 4 - n. This line used to read 
	 *	    ch &= ~ (mask >> (25 - (n << 2) - n)) 
	 * but now we simply use a table for speed. */

	ch &= uchmask[n];

	/* if there's no overlap (and optionally: no code point error), emit
	 * the decoded value (may break if no more room) and continue with
	 * the next sequence. */

	if ((n == 0 || ch > uchmask[n - 1])
#ifdef CHECKRANGE
	    && (ch < 0xd800 || (ch >= 0xe000 && ch < 0x10ffff))
#endif
	) { 
	    EMIT(ch);
	    continue;
	}

	/* otherwise, we have an overlap or code point range error; back up
	   to the very first octet of the sequence, emit an error, and if we
	   should continue, do so at the octet following it. */

	i -= n + 1; 
	EMITERR;
	i++;
    }
    if (ret_taken) *ret_taken = i - (unsigned char *)s;
    return ret;
}

